import tensorflow as tf
import numpy as np

# gpus = tf.config.experimental.list_physical_devices('GPU')

# if gpus:
#     try:
#         for gpu in gpus:
#             tf.config.experimental.set_memory_growth(gpu, True)
#     except RuntimeError as e:
#         print(e)


def grad(y, x):
    V = tf.keras.layers.Lambda(lambda z: tf.gradients(z[0], z[1]), output_shape=[1])(
        [y, x]
    )
    return V


def grad_cam_test(model, x, some_variable):
    fixed_input = model.inputs
    fixed_output = tf.keras.layers.Lambda(
        lambda z: tf.keras.backend.gradients(z[0], z[1]), output_shape=[2]
    )([model.inputs[0], model.outputs[0]])

    grad_model = tf.keras.Model(inputs=fixed_input, outputs=fixed_output)

    return grad_model.predict(x)


def run_test():
    # Generate sample model
    x = tf.keras.Input(shape=(2))
    y = tf.keras.layers.Dense(2)(x)
    model = tf.keras.Model(inputs=x, outputs=y)
    target = np.array([[1.0, 2.0]], dtype=np.float32)

    # Calculate gradient using numpy array
    input_numpy = np.array([[0.0, 0.0]])
    grad_output_numpy = grad_cam_test(model, input_numpy, target)
    print(f"numpy: {grad_output_numpy}")

    # Calculate gradient using tf.Variable
    input_variable = tf.constant([[0.0, 0.0]])
    grad_output_variable = grad_cam_test(model, input_variable, target)
    print(f"variable: {grad_output_variable}")


run_test()



import sys
from typing_extensions import Required

import click
import warnings
import os
import deepdanbooru as dd
import tensorflow.lite as tflite

__version__ = "1.0.1"


@click.version_option(prog_name="DeepDanbooru", version=__version__)
@click.group()
def main():
    """
    AI based multi-label girl image classification system, implemented by using TensorFlow.
    """
    pass


@main.command("create-project")
@click.argument(
    "project_path",
    type=click.Path(exists=False, resolve_path=True, file_okay=False, dir_okay=True),
)
def create_project(project_path):
    dd.commands.create_project(project_path)


@main.command("download-tags")
@click.option("--limit", default=10000, help="Limit for each category tag count.")
@click.option("--minimum-post-count", default=500, help="Minimum post count for tag.")
@click.option("--overwrite", help="Overwrite tags if exists.", is_flag=True)
@click.option("--username", help="Danbooru username for authentication.", required=True)
@click.option("--api-key", help="Danbooru API key for authentication.", required=True)
@click.argument(
    "path",
    type=click.Path(exists=False, resolve_path=True, file_okay=False, dir_okay=True),
)
def download_tags(path, limit, minimum_post_count, overwrite, username, api_key):
    dd.commands.download_tags(
        path, limit, minimum_post_count, overwrite, username, api_key
    )


@main.command("make-training-database")
@click.argument(
    "source_path",
    type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=False),
    nargs=1,
    required=True,
)
@click.argument(
    "output_path",
    type=click.Path(exists=False, resolve_path=True, file_okay=True, dir_okay=False),
    nargs=1,
    required=True,
)
@click.option(
    "--start-id",
    default=1,
    help="Start id.",
)
@click.option("--end-id", default=sys.maxsize, help="End id.")
@click.option("--use-deleted", help="Use deleted posts.", is_flag=True)
@click.option(
    "--chunk-size", default=5000000, help="Chunk size for internal processing."
)
@click.option("--overwrite", help="Overwrite tags if exists.", is_flag=True)
@click.option(
    "--vacuum", help="Execute VACUUM command after making database.", is_flag=True
)
def make_training_database(
    source_path,
    output_path,
    start_id,
    end_id,
    use_deleted,
    chunk_size,
    overwrite,
    vacuum,
):
    dd.commands.make_training_database(
        source_path,
        output_path,
        start_id,
        end_id,
        use_deleted,
        chunk_size,
        overwrite,
        vacuum,
    )


@main.command("train-project")
@click.argument(
    "project_path",
    type=click.Path(exists=True, resolve_path=True, file_okay=False, dir_okay=True),
)
@click.option(
    "--source-model",
    type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=False),
)
def train_project(project_path, source_model):
    dd.commands.train_project(project_path, source_model)


@main.command(
    "evaluate-project",
    help="Evaluate the project. If the target path is folder, it evaulates all images recursively.",
)
@click.argument(
    "project_path",
    type=click.Path(exists=True, resolve_path=True, file_okay=False, dir_okay=True),
)
@click.argument(
    "target_path",
    type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=True),
)
@click.option("--threshold", help="Threshold for tag estimation.", default=0.5)
def evaluate_project(project_path, target_path, threshold):
    dd.commands.evaluate_project(project_path, target_path, threshold)


@main.command(
    "grad-cam", help="Experimental feature. Calculate activation map using Grad-CAM."
)
@click.argument(
    "project_path",
    type=click.Path(exists=True, resolve_path=True, file_okay=False, dir_okay=True),
)
@click.argument(
    "target_path",
    type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=True),
)
@click.argument(
    "output_path",
    type=click.Path(resolve_path=True, file_okay=False, dir_okay=True),
    default=".",
)
@click.option("--threshold", help="Threshold for tag estimation.", default=0.5)
def grad_cam(project_path, target_path, output_path, threshold):
    dd.commands.grad_cam(project_path, target_path, output_path, threshold)


@main.command("evaluate", help="Evaluate model by estimating image tag.")
@click.argument(
    "target_paths",
    nargs=-1,
    type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=True),
)
@click.option(
    "--project-path",
    type=click.Path(exists=True, resolve_path=True, file_okay=False, dir_okay=True),
    help="Project path. If you want to use specific model and tags, use --model-path and --tags-path options.",
)
@click.option(
    "--model-path",
    type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=False),
)
@click.option(
    "--tags-path",
    type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=False),
)
@click.option("--threshold", default=0.5)
@click.option("--allow-gpu", default=False, is_flag=True)
@click.option("--compile/--no-compile", "compile_model", default=False)
@click.option(
    "--allow-folder",
    default=False,
    is_flag=True,
    help="If this option is enabled, TARGET_PATHS can be folder path and all images (using --folder-filters) in that folder is estimated recursively. If there are file and folder which has same name, the file is skipped and only folder is used.",
)
@click.option(
    "--save-txt",
    default=False,
    is_flag=True,
    help="Enable this option to save tags to a txt file with the same filename.",
)
@click.option(
    "--folder-filters",
    default="*.[Pp][Nn][Gg],*.[Jj][Pp][Gg],*.[Jj][Pp][Ee][Gg],*.[Gg][Ii][Ff]",
    help="Glob pattern for searching image files in folder. You can specify multiple patterns by separating comma. This is used when --allow-folder is enabled. Default:*.[Pp][Nn][Gg],*.[Jj][Pp][Gg],*.[Jj][Pp][Ee][Gg],*.[Gg][Ii][Ff]",
)
@click.option("--verbose", default=False, is_flag=True)
def evaluate(
    target_paths, # I guess its this one
    project_path,
    model_path,
    tags_path,
    threshold,
    allow_gpu,
    compile_model,
    allow_folder,
    save_txt,
    folder_filters,
    verbose,
):
    if verbose:
        warnings.filterwarnings("always")
        os.environ['TF_CPP_MIN_LOG_LEVEL'] = '0'
    dd.commands.evaluate(
        target_paths,
        project_path,
        model_path,
        tags_path,
        threshold,
        allow_gpu,
        compile_model,
        allow_folder,
        save_txt,
        folder_filters,
        verbose,
    )

@main.command("conv2tflite", help="Convert saved model into tflite model.")
@click.option(
    "--project-path",
    type=click.Path(exists=True, resolve_path=True, file_okay=False, dir_okay=True),
    help="Project path. If you want to use specific model and tags, use --model-path and --tags-path options.",
)
@click.option(
    "--model-path",
    type=click.Path(exists=True, resolve_path=True, file_okay=True, dir_okay=False),
)
@click.option(
    "--save-path",
    type=click.Path(resolve_path=True, file_okay=True, dir_okay=False),
)
@click.option("--optimize-default", default=True, is_flag=True)
@click.option("--optimize-experimental-sparsity", default=False, is_flag=True)
@click.option("--verbose", default=False, is_flag=True)
def conv2tflite(project_path, model_path, save_path, optimize_default, optimize_experimental_sparsity, verbose):
    if verbose:
        warnings.filterwarnings("always")
        os.environ['TF_CPP_MIN_LOG_LEVEL'] = '0'
    if not optimize_default and not optimize_experimental_sparsity:
        raise Exception("optimization method must be specified")
    op = []
    if optimize_default: op = [tflite.Optimize.DEFAULT]
    if optimize_experimental_sparsity: op.append(tflite.Optimize.EXPERIMENTAL_SPARSITY)
    dd.commands.convert_to_tflite_from_from_saved_model(project_path, model_path, save_path, op, verbose=verbose)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import importlib
from unittest import mock

import numpy
import pytest
import six
from click.testing import CliRunner
from PIL import Image


def test_import():
    import deepdanbooru


@pytest.mark.parametrize('func_name', ['main', 'evaluate'])
def test_help(func_name):
    mod = importlib.import_module('deepdanbooru.__main__')
    runner = CliRunner()
    result = runner.invoke(getattr(mod, func_name), ['--help'])
    assert result.exit_code == 0
    assert result.output


@pytest.fixture
def packages():
    with open('requirements.txt') as f:
        return f.read().splitlines()


def test_package_setup(packages):
    import setuptools
    with mock.patch('setuptools.setup'):
        import setup
        setup_pkgs = setup.install_requires
        tensorflow_pkg = setup.tensorflow_pkg
    assert setup_pkgs == list(
        filter(lambda x: not x.startswith('tensorflow'), packages))
    assert list(
        filter(lambda x: x.startswith('tensorflow'), packages)
    ) == [tensorflow_pkg]


def test_readme_pkg(packages):
    with open('README.md') as f:
        text = f.read()
    line_start = 'Following packages are need to be installed.'
    line_end = '\n\n'
    pkg_text = text.split(line_start)[1].rsplit(line_end)[0].strip()
    readme_pkgs = list(map(lambda x: x.split(' ', 1)[1], pkg_text.splitlines()))

    assert list(
        sorted(filter(lambda x: not x.startswith('tensorflow'), readme_pkgs))
    ) == list(
        sorted(filter(lambda x: not x.startswith('tensorflow'), packages))
    )

    assert list(
        filter(lambda x: x.startswith('tensorflow'), packages)
    ) == list(
        filter(lambda x: x.startswith('tensorflow'), readme_pkgs)
    )


@pytest.mark.parametrize('use_bytes_io', [True, False])
def test_load_image_for_evaluate(tmp_path, use_bytes_io):
    image_path = tmp_path / 'test.jpg'
    image = Image.new('RGB', (300, 300), color='red')
    image.save(image_path)
    from deepdanbooru.data import load_image_for_evaluate
    if not use_bytes_io:
        res = load_image_for_evaluate(image_path.as_posix(), 299, 299)
    else:
        with open(image_path.as_posix(), 'rb') as f:
            image_input = six.BytesIO(f.read())
        res = load_image_for_evaluate(image_input, 299, 299)
    assert isinstance(res, numpy.ndarray)
    assert res.shape == (299, 299, 3)



import warnings
warnings.filterwarnings("ignore")

import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import deepdanbooru.commands
import deepdanbooru.data
import deepdanbooru.extra
import deepdanbooru.image
import deepdanbooru.io
import deepdanbooru.model
import deepdanbooru.project
import deepdanbooru.train
